from typing import cast

from .client import GraphAPIClient
from .dataclasses import PageWebhookSubscription


class PageClient(GraphAPIClient):
    """
    Extends GraphAPIClient to include helpers for common Page-based functionality
    """

    def get_subscribed_apps(self) -> list[PageWebhookSubscription]:
        """
        Fetches a list of webhook subscriptions the page is subscribed to
        :return: List of parsed webhook subscription states
        """
        result = self.get('me/subscribed_apps')
        return [PageWebhookSubscription(**cast(dict, s)) for s in result.data]

    def subscribe_to_webhook(self, fields: list[str]) -> bool:
        """
        Subscribes a Page to the Meta webhook for the supplied fields
        :param fields: Webhook fields to subscribe the page to
        :return: Bool value indicating if the subscription was established
        """
        result = self.post('me/subscribed_apps', data={'subscribed_fields': ','.join(fields)})
        return cast(dict, result.data).get('success', False)



from __future__ import annotations

import logging
from types import TracebackType
from typing import Any, Final, Iterator, Optional, Type, cast

import backoff
from requests import Response, Session

from .constants import GRAPH_API_URL, GRAPH_API_VERSIONS
from .dataclasses import (
    AppUsageDetails,
    GraphAPIResponse,
    MarketingAPIThrottleInsights,
    TokenDebugPayload,
)
from .error_code import GraphAPICommonErrorCode
from .exceptions import (
    GraphAPIApplicationError,
    GraphAPIError,
    GraphAPIServiceError,
    GraphAPITokenError,
    GraphAPIUsageError,
    InvalidAccessToken,
    InvalidGraphAPIVersion,
)
from .typings import ErrorCodeExceptionMap, GraphAPIQueryResult, JSONTypeSimple

logger = logging.getLogger(__name__)


class GraphAPIClient:
    """
    A small client built to interact with the Facebook social graph:
    https://developers.facebook.com/docs/graph-api/overview

    This is currently built minimally for distribution in Dash Hudson
    services where JSON-based requests need to be handled.

    The following functionality is currently unsupported when
    comparing to the official facebook-sdk:
        - HMAC authentication
        - batch request handling
        - file uploading
        - generating oauth redirect urls

    For now API access is provisioned through access tokens,
    if you are unfamiliar with how this works see the following:
    https://developers.facebook.com/docs/facebook-login
    """

    DEFAULT_CODE_EXCEPTION_MAP: Final[ErrorCodeExceptionMap] = {
        (GraphAPICommonErrorCode.API_UNKNOWN.value, None): GraphAPIServiceError,
        (GraphAPICommonErrorCode.API_METHOD.value, None): GraphAPIServiceError,
        (GraphAPICommonErrorCode.API_PERMISSION_DENIED.value, None): GraphAPIApplicationError,
        (GraphAPICommonErrorCode.APPLICATION_BLOCKED_TEMP.value, None): GraphAPIApplicationError,
        (GraphAPICommonErrorCode.API_SESSION.value, None): GraphAPITokenError,
        (GraphAPICommonErrorCode.ACCESS_TOKEN_EXPIRED.value, None): GraphAPITokenError,
        (GraphAPICommonErrorCode.APPLICATION_LIMIT_REACHED.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.API_TOO_MANY_CALLS.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.PAGE_RATE_LIMIT_REACHED.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.CUSTOM_RATE_LIMIT_REACHED.value, None): GraphAPIUsageError,
    }

    def __init__(
        self,
        access_token: str,
        version: str,
        global_timeout: Optional[int] = None,
        params_to_mask: Optional[list[str]] = None,
        retry_params: Optional[dict] = None,
        disable_logger: Optional[bool] = False,
        code_exception_map: Optional[ErrorCodeExceptionMap] = None,
    ) -> None:
        """
        Initialize the API client
        :param access_token: An access token provisioned through Facebook login
        :param version: The Graph API version to use (ex: 12.0)
        :param global_timeout: A global request timeout to set
        :param params_to_mask: A list of query parameter names to mask when formatting
            exception messages
        :param disable_logger Disables exception logging if truthy
        :param retry_config Params for https://github.com/litl/backoff#backoffon_exception
        :param code_exception_map: A an error code -> exception map / configuration
        """
        if not access_token or not isinstance(access_token, str):
            raise InvalidAccessToken
        version = (
            version[1:] if isinstance(version, str) and version.lower().startswith('v') else version
        )
        if version not in GRAPH_API_VERSIONS:
            raise InvalidGraphAPIVersion(version)

        self.version = f'v{version}'
        self.global_timeout = global_timeout
        self.params_to_mask = params_to_mask
        self.disable_logger = disable_logger
        # Defaulting to max_tries=0 disables retrying by default
        self.retry_params = retry_params or {'exception': tuple(), 'max_tries': 0}

        self.code_exception_map = self.DEFAULT_CODE_EXCEPTION_MAP
        if code_exception_map:
            self.code_exception_map = {
                **self.DEFAULT_CODE_EXCEPTION_MAP,
                **code_exception_map,
            }

        self._access_token = access_token
        self._session = Session()
        self._session.params = {'access_token': self._access_token}

    def get(
        self,
        path: str,
        params: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
    ) -> GraphAPIResponse:
        """
        :param path: A path pointing to an edge or node
            (ex: /<page_id>/conversations)
        Performs a GET request to the Graph API
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='GET',
            path=path,
            params=params,
            timeout=timeout,
            retry_params=retry_params,
        )

    def post(
        self,
        path: str,
        data: Any,
        params: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
    ) -> GraphAPIResponse:
        """
        Performs a POST request to the Graph API
        :param path: A path pointing to an edge or node
            (ex: /<page_id>/conversations | /<page_id>)
        :param data: The request body to be included
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='POST',
            path=path,
            params=params,
            data=data,
            timeout=timeout,
            retry_params=retry_params,
        )

    def delete(
        self,
        path: str,
        params: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
    ) -> GraphAPIResponse:
        """
        Performs a DELETE request to the Graph API
        :param path: A path pointing to a node
            (ex: /<video_id>)
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='DELETE',
            path=path,
            params=params,
            timeout=timeout,
            retry_params=retry_params,
        )

    def debug_token(self, client_id: str, client_secret: str) -> TokenDebugPayload:
        """
        Method for debugging the contained access token
        :param client_id: Meta developer app client ID
        :param client_secret: Meta developer app client secret
        :return: A parsed token debug payload
        """
        result = self.get(
            'debug_token',
            params={
                'access_token': f'{client_id}|{client_secret}',
                'input_token': self._access_token,
            },
        )
        return TokenDebugPayload.from_source_data(cast(dict, result.data))

    def _do_request(
        self,
        method: str,
        path: str,
        params: Optional[Any] = None,
        data: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
    ) -> GraphAPIResponse:
        """
        Handle Graph API requests. Raise if error body detected and lets ambiguous network
        errors propagate.
        :param method: The HTTP request method
        :param path: A path pointing to an edge or node (ex: /<page_id>/conversations)
        :param params: Query parameters to be included with the request
        :param data: The request body to be included
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """

        @backoff.on_exception(backoff.expo, **(retry_params or self.retry_params))
        def _retry_parameterizer() -> GraphAPIResponse:
            response = self._session.request(
                method=method,
                url=f'{GRAPH_API_URL}/{self.version}/{path}',
                params=params,
                data=data,
                timeout=timeout or self.global_timeout,
            )
            result, paging = self._parse_response_body_or_raise(response)

            return GraphAPIResponse(
                app_usage_details=AppUsageDetails.from_header(response),
                marketing_api_throttle_insights=MarketingAPIThrottleInsights.from_header(response),
                data=result,
                paging=paging,
            )

        return _retry_parameterizer()

    def _parse_response_body_or_raise(
        self, response: Response
    ) -> tuple[GraphAPIQueryResult, Optional[JSONTypeSimple]]:
        """
        Parse Graph API response body and raise if error details present
        :param response: A response from the Graph API
        :return: Parsed request body and optional paging params
        """
        response_body = response.json()
        error_details = response_body.get('error')
        if error_details:
            code_key = (error_details.get('code'), error_details.get('error_subcode'))
            # Raise a specific exception if a code mapping is set, custom exceptions take priority
            exc_type = self.code_exception_map.get(code_key, GraphAPIError)
            # Log & raise default GraphAPIError if no mapping was found
            exc = exc_type(
                response=response,
                error_details=error_details,
                params_to_mask=self.params_to_mask,
            )
            if not self.disable_logger:
                logger.error(str(exc))
            raise exc
        # If 'data' is present, it means the result is a list of graph nodes and may have
        # paging params as well
        if 'data' in response_body:
            return response_body['data'], response_body.get('paging')
        # If not, the response body is a single graph node without paging params
        return response_body, None

    def _get_paginated_results(self, path: str, params: JSONTypeSimple) -> Iterator[JSONTypeSimple]:
        """
        Simple helper used for iterating over paged-results
        TODO: Could be exposed for general use but should implement complete feature-set
        TODO: If exposed for public use, consider return type a bit
        :param path: Path to pass to self.get
        :param params: Params to pass to self.get
        :return: Iterator that yields paged results
        """
        while True:
            result = self.get(path=path, params=params)
            items = cast(list[JSONTypeSimple], result.data)
            for i in items:
                yield i

            if not result.after_cursor:
                break
            params['after'] = result.after_cursor

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._session.close()



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Union

from requests import Response

from .helpers import deserialize_json_header, posix_2_datetime
from .typings import GraphAPIQueryResult, JSONTypeSimple


@dataclass(frozen=True)
class TokenDebugPayload:
    """
    Encapsulates a token debug payload from the Graph API
    See: https://developers.facebook.com/docs/facebook-login/guides/%20access-tokens/debugging
    """

    app_id: str
    type: str
    application: str
    data_access_expires_at: datetime
    expires_at: datetime
    is_valid: bool
    user_id: str
    scopes: list[str]
    granular_scopes: list[GranularScope]
    issued_at: Optional[datetime] = None  # Only present if token is long-lived

    @classmethod
    def from_source_data(cls, source_data: JSONTypeSimple) -> TokenDebugPayload:
        date_fields: JSONTypeSimple = {}
        for k in ['data_access_expires_at', 'expires_at', 'issued_at']:
            timestamp = source_data.pop(k, None)
            if timestamp:
                date_fields[k] = posix_2_datetime(timestamp)

        granular_scopes = source_data.pop('granular_scopes', [])
        if granular_scopes:
            granular_scopes = [GranularScope.from_source_data(s) for s in granular_scopes]

        return TokenDebugPayload(**source_data, **date_fields, granular_scopes=granular_scopes)


@dataclass(frozen=True)
class GranularScope:
    """
    Encapsulates a granular scope state returned via token debug payload
    See: https://developers.facebook.com/docs/facebook-login/guides/%20access-tokens/debugging
    """

    scope: str
    target_ids: Optional[list[str]] = None

    def contains_target_id(self, target_id: Union[int, str]) -> bool:
        if not self.target_ids:
            return True
        return str(target_id) in self.target_ids

    @classmethod
    def from_source_data(cls, source_data: JSONTypeSimple) -> GranularScope:
        return GranularScope(scope=source_data['scope'], target_ids=source_data.get('target_ids'))


@dataclass(frozen=True)
class UserScope:
    """
    Encapsulates a user state object returned by /me/permissions
    See: https://developers.facebook.com/docs/graph-api/reference/user/permissions/
    """

    permission: str
    status: str

    @property
    def is_granted(self) -> bool:
        return self.status == 'granted'


@dataclass(frozen=True)
class ConnectedPage:
    """
    Encapsulates a partial Page state
    See: https://developers.facebook.com/docs/graph-api/reference/page/
    """

    id: str
    access_token: str
    instagram_account_id: Optional[int] = None
    instagram_business_id: Optional[str] = None
    tasks: Optional[list[str]] = None

    def matches_id_set(self, page_id: int, instagram_account_id: Optional[int]) -> bool:
        page_id_matches = self.id == page_id
        if not instagram_account_id:
            return page_id_matches
        return page_id_matches and self.instagram_account_id == instagram_account_id

    @classmethod
    def from_source_data(cls, source_data: JSONTypeSimple) -> ConnectedPage:
        linked_instagram_account = source_data.get('instagram_business_account', {})
        return ConnectedPage(
            id=source_data['id'],
            access_token=source_data['access_token'],
            instagram_account_id=linked_instagram_account.get('ig_id'),
            instagram_business_id=linked_instagram_account.get('id'),
            tasks=source_data.get('tasks', []),
        )


@dataclass(frozen=True)
class PageWebhookSubscription:
    """
    Encapsulates a Webhook subscription state for a page
    See: https://developers.facebook.com/docs/graph-api/reference/page/subscribed_apps
    """

    category: str
    link: str
    name: str
    id: str
    subscribed_fields: list[str]


@dataclass(frozen=True)
class AppUsageDetails:
    """
    Encapsulates stats from X-App-Usage header:
    https://developers.facebook.com/docs/graph-api/overview/rate-limiting#headers
    """

    call_count: int
    total_time: int
    total_cputime: int

    @classmethod
    def from_header(cls, res: Response) -> AppUsageDetails:
        app_usage_dict = deserialize_json_header(res=res, header_name='X-App-Usage')
        return cls(
            call_count=app_usage_dict.get('call_count', 0),
            total_time=app_usage_dict.get('total_time', 0),
            total_cputime=app_usage_dict.get('total_cputime', 0),
        )


@dataclass(frozen=True)
class MarketingAPIThrottleInsights:
    """
    Encapsulates stats from X-Fb-Ads-Insights-Throttle header:
    https://developers.facebook.com/docs/marketing-api/insights/best-practices/#insightscallload
    """

    app_id_util_pct: float
    acc_id_util_pct: float
    ads_api_access_tier: str

    @classmethod
    def from_header(cls, res: Response) -> MarketingAPIThrottleInsights:
        throttle_insights_dict = deserialize_json_header(
            res=res, header_name='X-Fb-Ads-Insights-Throttle'
        )
        return cls(
            app_id_util_pct=throttle_insights_dict.get('app_id_util_pct', 0.0),
            acc_id_util_pct=throttle_insights_dict.get('acc_id_util_pct', 0.0),
            ads_api_access_tier=throttle_insights_dict.get('ads_api_access_tier', ''),
        )


@dataclass
class GraphAPIResponse:
    """
    Encapsulates a Graph API response payload with parsed app usage headers
    """

    app_usage_details: AppUsageDetails
    marketing_api_throttle_insights: MarketingAPIThrottleInsights
    data: GraphAPIQueryResult
    paging: Optional[JSONTypeSimple] = None

    @property
    def is_empty(self) -> bool:
        return not self.data

    @property
    def is_list(self) -> bool:
        return isinstance(self.data, list)

    @property
    def is_dict(self) -> bool:
        return isinstance(self.data, dict)

    @property
    def before_cursor(self) -> Optional[str]:
        return self.cursors.get('before')

    @property
    def after_cursor(self) -> Optional[str]:
        return self.cursors.get('after')

    @property
    def cursors(self) -> JSONTypeSimple:
        return self.paging.get('cursors', {}) if self.paging else {}



from .client import GraphAPIClient
from .dataclasses import (
    AppUsageDetails,
    ConnectedPage,
    GranularScope,
    GraphAPIResponse,
    PageWebhookSubscription,
    TokenDebugPayload,
    UserScope,
)
from .error_code import GraphAPICommonErrorCode
from .exceptions import (
    GraphAPIApplicationError,
    GraphAPIError,
    GraphAPIServiceError,
    GraphAPITokenError,
    GraphAPIUsageError,
    InvalidAccessToken,
    InvalidGraphAPIVersion,
)
from .helpers import FieldConfig, build_field_config_list, format_fields_str
from .page_client import PageClient
from .user_client import UserClient

__all__ = [
    'GraphAPIClient',
    'AppUsageDetails',
    'ConnectedPage',
    'GranularScope',
    'GraphAPIResponse',
    'PageWebhookSubscription',
    'TokenDebugPayload',
    'UserScope',
    'GraphAPICommonErrorCode',
    'GraphAPIApplicationError',
    'GraphAPIError',
    'GraphAPIServiceError',
    'GraphAPITokenError',
    'GraphAPIUsageError',
    'InvalidAccessToken',
    'InvalidGraphAPIVersion',
    'FieldConfig',
    'build_field_config_list',
    'format_fields_str',
    'PageClient',
    'UserClient',
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dh_facebook_client']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.11.1,<2.0.0', 'requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'dh-facebook-client',
    'version': '0.2.9',
    'description': 'Simple client for interacting with the Facebook Graph API',
    'long_description': '# dh-facebook-client\n',
    'author': 'pchisholm',
    'author_email': 'chisholm.p@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Final, Optional, cast

from .client import GraphAPIClient
from .dataclasses import ConnectedPage, UserScope
from .helpers import first_true, format_fields_str

CONNECTED_PAGES_DEFAULT_LIMIT: Final = 100
CONNECTED_PAGES_FIELDS: Final = format_fields_str(
    [
        {'field_name': 'access_token'},
        {'field_name': 'tasks'},
        {
            'field_name': 'instagram_business_account',
            'sub_fields': [
                {
                    'field_name': 'ig_id',
                },
            ],
        },
    ],
)


class UserClient(GraphAPIClient):
    """
    Extends GraphAPIClient to include helpers for common User-based functionality
    """

    def get_user_scopes(self) -> list[UserScope]:
        """
        Return a list of scopes requested by the developer app at the User-level
        :return: List of UserScope(s)
        """
        return [UserScope(**cast(dict, s)) for s in self.get(path='me/permissions').data]

    def find_connected_page(
        self,
        page_id: int,
        instagram_account_id: Optional[int] = None,
        limit: Optional[int] = CONNECTED_PAGES_DEFAULT_LIMIT,
    ) -> Optional[ConnectedPage]:
        """
        Find a Page connected to a User by either Page ID and / or Instagram Account ID
        :param page_id: The Page ID to match
        :param instagram_account_id: Optional Instagram Account ID to match
        :param limit: Number of partial pages to return
        :return: Optional ConnectedPage partial
        """
        return first_true(
            (
                ConnectedPage.from_source_data(cast(dict, p))
                for p in self._get_paginated_results(
                    'me/accounts', params={'fields': CONNECTED_PAGES_FIELDS, 'limit': limit}
                )
            ),
            lambda p: p.matches_id_set(page_id, instagram_account_id),
        )
